using Microsoft.SqlServer.Management.Smo;
using Microsoft.SqlServer.Management.Common;
using ShuffleText.Highlight.Express;
using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;

namespace DatabaseObjectSearcher
{
    public class DatabaseObjectSearcher
    {
        SqlConnectionInfo connectionInfo;
        Server server;
        private object searchLock = new object();
        SearchableDictionary<DatabaseSearchResult> databaseObjectDictionary = new SearchableDictionary<DatabaseSearchResult>();
        private static Regex splitRegEx = new Regex("([A-Z]{1,1})");

        public DatabaseObjectSearcher(SqlConnectionInfo connectionInfo)
        {
            this.connectionInfo = connectionInfo;
            server = new Server(new ServerConnection(connectionInfo));

            // these give you a HUGE perf win with SMO - it pre-fetches these, rather than having to make another call to SQL Server to get this value
            server.SetDefaultInitFields(typeof(StoredProcedure), "IsSystemObject");
            server.SetDefaultInitFields(typeof(View), "IsSystemObject");
            server.SetDefaultInitFields(typeof(Table), "IsSystemObject");
            server.SetDefaultInitFields(typeof(Database), "IsSystemObject");
            server.SetDefaultInitFields(typeof(Database), "IsAccessible");

            System.Diagnostics.Debug.WriteLine("new searcher created");
        }

        public List<DatabaseSearchResult> FindMatchingObjects(string name)
        {
            List<DatabaseSearchResult> matches = new List<DatabaseSearchResult>();
            IList<KeyValuePair<DatabaseSearchResult, string>> matchList = null;
            if (Monitor.TryEnter(searchLock, 1000)) // times out after 1 second
            {
                try
                {
                    matchList = databaseObjectDictionary.Search(name, 50);
                }
                finally
                {
                    Monitor.Exit(searchLock);
                }
            }

            if (matchList != null)
            {
                foreach (KeyValuePair<DatabaseSearchResult, string> dbObject in matchList)
                {
                    matches.Add(dbObject.Key);
                }
            }

            return matches;
        }

        internal void BuildDBObjectDictionary(object ignored)
        {
            lock (searchLock)
            {
                foreach (Database d in server.Databases)
                {
                    
                    if (!d.IsSystemObject && d.IsAccessible) // even with pre-fetching this is still slow if there are thousands of databases on a remote server with a slow connection
                    {
                        databaseObjectDictionary.Add(new DatabaseSearchResult(d, connectionInfo), d.Name);
                                             
                        try
                        {
                            foreach (Table t in d.Tables)
                            {
                                if (!t.IsSystemObject)
                                {
                                    databaseObjectDictionary.Add(new DatabaseSearchResult(t, connectionInfo), GetSearchStringForDbObjectName(t.Name));
                                }
                            }

                            foreach (StoredProcedure p in d.StoredProcedures)
                            {
                                if (!p.IsSystemObject)
                                {
                                    databaseObjectDictionary.Add(new DatabaseSearchResult(p, connectionInfo), GetSearchStringForDbObjectName(p.Name));
                                }
                            }

                            foreach (View v in d.Views)
                            {
                                if (!v.IsSystemObject)
                                {
                                    databaseObjectDictionary.Add(new DatabaseSearchResult(v, connectionInfo), GetSearchStringForDbObjectName(v.Name));
                                }
                            }
                        }
                        catch (ExecutionFailureException)
                        {
                            // this can get thrown for security reasons - probably need to swallow here

                        }
                        catch (EnumeratorException)
                        {
                            // this also seems to be a possible security-related exception - also swallow
                        }
                    }
                }
            }
        }

        private static string GetSearchStringForDbObjectName(string name)
        {
            return name + " " + name.Replace("_", " ") + " " + SplitCasedString(name);
        }

        private static string SplitCasedString(string input)
        {
            if (splitRegEx.IsMatch(input))
            {
                // split out a pascal-cased or camel-cased string in to "words" 
                // TODO - this won't handle acronyms in table names very well (for instance ADODotNet would get split to A D O Dot Net) 
                return splitRegEx.Replace(input, " $1");
            }
            else
            {
                return string.Empty;
            }
        }

    }
}
